@@ -1,10 +1,15 @@
 import datetime
-from datetime import timedelta
+from datetime import timedelta, timezone
+from functools import partial
 from numbers import Number
 
 import numpy as np
+from dateutil.tz import tzlocal
+from scipy import stats
 from scipy.signal import argrelextrema
 
+import Orange.data.util
+
 
 def _parse_args(args, kwargs, names, *defaults):
     assert len(defaults) == len(names)
@@ -183,7 +188,7 @@ def _significant_acf(corr, has_confint):
     return result
 
 
-def autocorrelation(x, *args, unbiased=True, nlags=None, fft=True, **kwargs):
+def autocorrelation(x, *args, nlags=None, fft=True, **kwargs):
     """
     Return autocorrelation function of signal `x`.
 
@@ -208,7 +213,7 @@ def autocorrelation(x, *args, unbiased=True, nlags=None, fft=True, **kwargs):
     from statsmodels.tsa.stattools import acf
     if nlags is None:
         nlags = int(.9 * len(x))
-    corr = acf(x, *args, unbiased=unbiased, nlags=nlags, fft=fft, **kwargs)
+    corr = acf(x, *args, nlags=nlags, fft=fft, **kwargs)
     return _significant_acf(corr, kwargs.get('alpha'))
 
 
@@ -222,7 +227,7 @@ def partial_autocorrelation(x, *args, nlags=None, method='ldb', **kwargs):
         A 1D signal.
     nlags: int
         The number of lags to calculate the correlation for
-        (default: min(600, len(x)))
+        (default: min(len(x)//2 - 1, len(x) - 1))
     args, kwargs
         As accepted by `statsmodels.tsa.stattools.pacf`.
 
@@ -235,7 +240,7 @@ def partial_autocorrelation(x, *args, nlags=None, method='ldb', **kwargs):
     """
     from statsmodels.tsa.stattools import pacf
     if nlags is None:
-        nlags = min(1000, len(x) - 1)
+        nlags = min(len(x)//2 - 1, len(x) - 1)
     corr = pacf(x, *args, nlags=nlags, method=method, **kwargs)
     return _significant_acf(corr, kwargs.get('alpha'))
 
@@ -263,6 +268,8 @@ def interpolate_timeseries(data, method='linear', multivariate=False):
     from Orange.data import Domain
     from orangecontrib.timeseries import Timeseries
 
+    assert method in ("linear", "cubic", "mean", "nearest")
+
     attrs = data.domain.attributes
     cvars = data.domain.class_vars
     metas = data.domain.metas
@@ -271,7 +278,7 @@ def interpolate_timeseries(data, method='linear', multivariate=False):
     M = data.metas.copy()
 
     # Interpolate discrete columns to mode/nearest value
-    _x = np.arange(1, 1 + len(data))
+    _x = Timeseries.from_data_table(data).time_values.astype(float)
     for A, vars in ((X, attrs),
                     (Y, cvars)):
         for i, var in enumerate(vars):
@@ -394,7 +401,7 @@ def seasonal_decompose(data, model='multiplicative', period=12, *, callback=None
     for var in data.domain.variables:
         decomposed = sm.tsa.seasonal_decompose(np.ravel(interp_data[:, var]),
                                                model=model,
-                                               freq=period)
+                                               period=period)
         adjusted = recomposition(decomposed.observed,
                                  decomposed.seasonal)
 
@@ -442,7 +449,7 @@ def granger_causality(data, max_lag=10, alpha=.05, *, callback=None):
     Returns
     -------
     res : list of lists
-        Each internal list is [lag, antecedent, consequent] where
+        Each internal list is [lag, p-value, antecedent, consequent] where
         lag is the minimum lag at which antecedent feature in data is
         Granger-causal for the consequent feature in data.
     """
@@ -457,90 +464,133 @@ def granger_causality(data, max_lag=10, alpha=.05, *, callback=None):
     domain = [var for var in data.domain.variables if var.is_continuous]
     res = []
 
+    step = 0
     for row_attr in domain:
         for col_attr in domain:
             if row_attr == col_attr or data.time_variable in (row_attr, col_attr):
                 continue
-            X = Table(Domain([], [], [col_attr, row_attr], data.domain), data).metas
+            X = Table.from_table(Domain([], [], [col_attr, row_attr], data.domain), data).metas
             try:
                 tests = grangercausalitytests(X, max_lag, verbose=False)
-                lag = next((lag for lag in range(1, 1 + max_lag)
-                            if tests[lag][0]['ssr_ftest'][1] < alpha), 0)
+                lag, p_val = next(
+                    (
+                        (lag, tests[lag][0]["ssr_ftest"][1])
+                        for lag in range(1, 1 + max_lag)
+                        if tests[lag][0]["ssr_ftest"][1] < alpha
+                    ),
+                    (None, None),
+                )
             except ValueError:
-                lag = 0
+                lag = p_val = None
             if lag:
-                res.append([lag, row_attr.name, col_attr.name])
+                res.append([lag, p_val, row_attr.name, col_attr.name])
+
+            step += 1
             if callback:
-                callback(1 / ((len(domain) - 1)**2 - len(domain)))
+                callback(step / ((len(domain) - 1) ** 2 - len(domain) + 1))
     return res
 
 
-def moving_transform(data, spec, fixed_wlen=0):
-    """
-    Return data transformed according to spec.
+def moving_sum(x, width, shift=1):
+    s = np.nancumsum(x)
+    return np.hstack((s[width - 1:width] - 0,
+                      s[shift + width - 1::shift]
+                      - s[shift - 1:-width:shift]))
 
-    Parameters
-    ----------
-    data : Timeseries
-        A table with features to transform.
-    spec : list of lists
-        A list of lists [feature:Variable, window_length:int, function:callable].
-    fixed_wlen : int
-        If not 0, then window_length in spec is disregarded and this length
-        is used. Also the windows don't shift by one but instead align
-        themselves side by side.
 
-    Returns
-    -------
-    transformed : Timeseries
-        A table of original data its transformations.
-    """
-    from itertools import chain
-    from Orange.data import ContinuousVariable, Domain
-    from orangecontrib.timeseries import Timeseries
-    from orangecontrib.timeseries.widgets.utils import available_name
-    from orangecontrib.timeseries.agg_funcs import Cumulative_sum, Cumulative_product
+def moving_count_nonzero(x, width, shift=1):
+    return moving_sum((x != 0) & np.isfinite(x), width, shift)
 
-    X = []
-    attrs = []
 
-    for var, wlen, func in spec:
-        col = np.ravel(data[:, var])
+def moving_count_defined(x, width, shift=1):
+    return moving_sum(np.isfinite(x), width, shift)
 
-        if fixed_wlen:
-            wlen = fixed_wlen
 
-        if func in (Cumulative_sum, Cumulative_product):
-            out = list(chain.from_iterable(func(col[i:i + wlen])
-                                           for i in range(0, len(col), wlen)))
-        else:
-            # In reverse cause lazy brain. Also prefer informative ends, not beginnings as much
-            col = col[::-1]
-            out = [func(col[i:i + wlen])
-                   for i in range(0, len(col), wlen if bool(fixed_wlen) else 1)]
-            out = out[::-1]
+def _windowed(x, width, shift):
+    if width > x.size:
+        return np.empty((0, 1))  # we need a 2d array, but 0 rows
+    return np.lib.stride_tricks.as_strided(
+        x,
+        shape=(1 + (x.size - width) // shift, width),
+        strides=(shift * x.strides[0], x.strides[0])
+    )
 
-        X.append(out)
 
-        template = '{} ({}; {})'.format(var.name, wlen, func.__name__.lower().replace('_', ' '))
-        name = available_name(data.domain, template)
-        attrs.append(ContinuousVariable(name))
+def windowed_func(func, x, width, shift):
+    return func(_windowed(x, width, shift), axis=1)
 
-    dataX, dataY, dataM = data.X, data.Y, data.metas
-    if fixed_wlen:
-        n = len(X[0])
-        dataX = dataX[::-1][::fixed_wlen][:n][::-1]
-        dataY = dataY[::-1][::fixed_wlen][:n][::-1]
-        dataM = dataM[::-1][::fixed_wlen][:n][::-1]
 
-    ts = Timeseries.from_numpy(Domain(data.domain.attributes + tuple(attrs),
-                                      data.domain.class_vars,
-                                      data.domain.metas),
-                               np.column_stack(
-                                   (dataX, np.column_stack(X))) if X else dataX,
-                               dataY, dataM)
-    ts.time_variable = data.time_variable
-    return ts
+def windowed_span(x, width, shift):
+    windows = _windowed(x, width, shift)
+    return np.nanmax(windows, axis=1) - np.nanmin(windows, axis=1)
+
+
+def _windowed_weighted(x, weights, shift):
+    xnans = np.isnan(x)
+    if not np.any(xnans):
+        return np.sum(_windowed(x, len(weights), shift) * weights, axis=1)
+
+    # Recompute weights for each line so that the total sum is the same
+    # after skipping the weights that correspond to nan
+    # If the sum of weights is 1, this is just "renormalization"
+    x = x.copy()
+    windows = _windowed(x, len(weights), shift)
+    nans = np.isnan(windows)
+    total_weight = np.sum(weights)
+    weights = np.repeat(weights[None, :], len(windows), axis=0)
+    weights[nans] = 0
+    x[xnans] = 0
+    weightsums = np.sum(weights, axis=1) / total_weight
+    no_data = weightsums == 0
+    weightsums[no_data] = 1
+    res = np.sum(windows * weights, axis=1) / weightsums
+    res[no_data] = np.nan
+    return res
+
+def windowed_linear_MA(x, width, shift):
+    weights = np.arange(1, width + 1, dtype=float)
+    weights /= np.sum(weights)
+    return _windowed_weighted(x, weights, shift)
+
+
+def windowed_exponential_MA(x, width, shift):
+    alpha = 2 / (width + 1.0)
+    weights = alpha * (1 - alpha) ** np.arange(width - 1, -1, -1)
+    weights /= np.sum(weights)
+    return _windowed_weighted(x, weights, shift)
+
+
+def windowed_cumsum(x, width, shift):
+    return np.nancumsum(x)[width - 1::shift]
+
+
+def windowed_cumprod(x, width, shift):
+    return np.nancumprod(x)[width - 1::shift]
+
+
+def windowed_mode(x, width, shift):
+    modes, counts = windowed_func(
+        partial(stats.mode, nan_policy='omit'),
+        x, width, shift)
+    modes = modes[:, 0]
+    if np.ma.isMaskedArray(modes):
+        # If counts == 0, all values were nan
+        modes = modes.data
+        modes[counts[:, 0] == 0] = np.nan
+    return modes
+
+def windowed_harmonic_mean(x, width, shift):
+    windows = _windowed(x, width, shift)
+    try:
+        return stats.hmean(windows, axis=1)
+    except ValueError:
+        r = np.full(len(windows), np.nan)
+        for i, window in enumerate(windows):
+            try:
+                r[i] = stats.hmean(window)
+            except ValueError:
+                pass
+        return r
 
 
 def model_evaluation(data, models, n_folds, forecast_steps, *, callback=None):
@@ -625,7 +675,7 @@ def model_evaluation(data, models, n_folds, forecast_steps, *, callback=None):
             if fittedvalues.ndim > 1:
                 fittedvalues = fittedvalues[..., 0]
         except Exception:
-            row = ['err'] * 7
+            row = ['err'] * 8
         else:
             row = _score_vector(model, true_y, fittedvalues)
         row[0] = row[0] + ' (in-sample)'
@@ -640,9 +690,13 @@ DAYS = 1461
 def timestamp(dt):
     try:
         ts = dt.timestamp()
-    except OverflowError:
-        k = (1970 - dt.year) / 4 + 1
-        ts = (dt + timedelta(days=k * DAYS)).timestamp() - k * SECONDS
+    except (OverflowError, OSError):
+        if not dt.tzinfo:
+            # treat datetime as in local timezone - workaround for astimezone
+            # which does not work on Windows for dates older than 1970
+            dt = dt.replace(tzinfo=tzlocal())
+        # compute timestamp manually
+        ts = (dt - datetime.datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds()
     return ts
 
 
@@ -654,3 +708,11 @@ def fromtimestamp(ts, tz=None):
         dt = datetime.datetime.fromtimestamp(ts + k * SECONDS, tz=tz) - \
              timedelta(days=k * DAYS)
     return dt
+
+
+def truncated_date(date, level):
+    kwargs = {unit: zeroed
+              for unit, zeroed  in (
+                  ("month", 1), ("day", 1),
+                  ("hour", 0), ("minute", 0), ("second", 0))[level:]}
+    return date.replace(**kwargs)
