@@ -11,25 +11,27 @@ from AnyQt.QtWidgets import (
     QListView,
     QVBoxLayout,
 )
-from AnyQt.QtGui import QIcon, QCloseEvent
+from AnyQt.QtGui import QIcon
 from AnyQt.QtCore import QSize, pyqtSignal
 
 from orangewidget.utils.widgetpreview import WidgetPreview
 from Orange.data import TimeVariable, Table, Variable
 from Orange.widgets import gui
 from Orange.widgets.widget import OWWidget, Input, AttributeList
-from Orange.widgets.utils import vartype
 from Orange.widgets.settings import (
     ContextSetting,
-    IncompatibleContext,
-    DomainContextHandler,
+    ContextHandler,
+    Context
 )
 from Orange.widgets.utils.itemmodels import VariableListModel
 
-from orangecontrib.timeseries import Timeseries
+from orangecontrib.timeseries import Timeseries, fromtimestamp
 from orangecontrib.timeseries.widgets.highcharts import Highchart
 
 
+PLOT_TYPE_ITEMS = ('line', 'step line', 'column', 'area', 'spline')
+
+
 class PlotConfigWidget(QWidget, gui.OWComponent):
     sigClosed = pyqtSignal(str, QWidget)
     sigLogarithmic = pyqtSignal(str, bool)
@@ -56,10 +58,10 @@ class PlotConfigWidget(QWidget, gui.OWComponent):
         self.setLayout(box)
 
         hbox = gui.hBox(self)
-        gui.comboBox(hbox, self, 'plot_type',
+        self.plot_type_cb = gui.comboBox(hbox, self, 'plot_type',
                      label='Type:',
                      orientation='horizontal',
-                     items=('line', 'step line', 'column', 'area', 'spline'),
+                     items=PLOT_TYPE_ITEMS,
                      sendSelectedValue=True,
                      callback=lambda: self.sigType.emit(ax, self.plot_type))
         gui.rubber(hbox)
@@ -150,9 +152,13 @@ class PlotConfigWidget(QWidget, gui.OWComponent):
         self.is_logarithmic = is_log
         self.on_logarithmic()
 
+    def set_plot_type(self, plot_type: str) -> None:
+        self.plot_type = plot_type
+        self.plot_type_cb.setCurrentText(plot_type)
+        self.sigType.emit(self.ax, self.plot_type)
+
 
 class Highstock(Highchart):
-
     def __init__(self, parent, *args, **kwargs):
         super().__init__(parent, *args,
                          yAxis_lineWidth=2,
@@ -405,40 +411,41 @@ class Highstock(Highchart):
         ''' % locals())
 
 
-class LineChartContextHandler(DomainContextHandler):
+class LineChartContextHandler(ContextHandler):
     """
     Context handler of line chart. The specifics of this widget is
     that `attrs` variable is list of lists and it is not handled with
     the DomainContextHandler.
     """
+    def new_context(self, group):
+        context = super().new_context()
+        context.features = group[1]
+        return context
 
-    def encode_setting(self, context, setting, value):
-        if setting.name == "attrs":
-            value = [
-                [(var.name, vartype(var)) for var in plot] for plot in value
-            ]
-        return super().encode_setting(context, setting, value)
-
-    def decode_setting(self, setting, value, domain=None, *_args):
-        decoded = super().decode_setting(setting, value, domain)
-        if setting.name == "attrs":
-            decoded = [[domain[name] for name, _ in plot] for plot in decoded]
-        return decoded
-
-    def match(self, context, _, attrs, metas):
-        if context.attributes == attrs and context.metas == metas:
-            return self.PERFECT_MATCH
-
+    def match(self, context, domain_features, *args):
         if "attrs" not in context.values:
             return self.NO_MATCH
 
-        all_vars = attrs.copy()
-        all_vars.update(metas)
-
-        value = set(y for x in context.values["attrs"] for y in x)
-        return sum(all_vars.get(attr) == vtype for attr, vtype in value) / len(
-            value
-        )
+        domain, features = domain_features
+        attrs = context.values["attrs"]
+        if features is not None:
+            # when features are present match features only - find context with
+            # same features
+            if context.features == features:
+                return self.PERFECT_MATCH
+            else:
+                return self.NO_MATCH
+        else:
+            # match if all selected attributes in domain
+            # when no features on input match just contexts with no features
+            values = set(y for x in attrs for y in x)
+            if (
+                context.features is None and len(values) > 0
+                and all(v in domain for v in values)
+            ):
+                return self.PERFECT_MATCH
+            else:
+                return self.NO_MATCH
 
 
 class OWLineChart(OWWidget):
@@ -453,8 +460,11 @@ class OWLineChart(OWWidget):
         forecast = Input("Forecast", Timeseries, multiple=True)
 
     settingsHandler = LineChartContextHandler()
-    attrs = ContextSetting([])  # Maps data.name -> [attrs]
-    is_logit = ContextSetting([])
+    attrs: List[List[str]] = ContextSetting([])
+    is_logit: List[bool] = ContextSetting([])
+    plot_type: List[str] = ContextSetting([])
+
+    settings_version = 2
 
     graph_name = 'chart'
 
@@ -508,7 +518,7 @@ class OWLineChart(OWWidget):
             self.resize(QSize(925, 635))
 
     @Inputs.time_series
-    def set_data(self, data):
+    def set_data(self, data: Table):
         # TODO: set xAxis resolution and tooltip time contents depending on
         # data.time_delta. See: http://imgur.com/yrnlgQz
 
@@ -523,20 +533,28 @@ class OWLineChart(OWWidget):
             return
 
         self.data = data = new_data
+        self.closeContext()
         if data is None:
             self.varmodel.clear()
             self.chart.clear()
             return
 
-        self.set_attributes()
-
-        if getattr(data.time_variable, 'utc_offset', False):
-            offset_minutes = data.time_variable.utc_offset.total_seconds() / 60
-            self.chart.evalJS(
-                'Highcharts.setOptions({global: {timezoneOffset: %d}});'
-                % -offset_minutes
-            )  # Why is this negative? It works.
-            self.chart.chart()
+        if getattr(data.time_variable, "timezone", False):
+            # get all datetime values in as timestamp
+            t_values = data.get_column_view(data.time_variable)[0]
+            # get all offsets (it is possible to have multiple offsets because
+            # of changes in history and changes to dst)
+            offsets = set(
+                fromtimestamp(s, tz=data.time_variable.timezone).utcoffset()
+                for s in t_values
+            ) - {None}
+            # set offset only when same offset for all datetimes
+            if len(offsets) == 1:
+                offset_minutes = offsets.pop().total_seconds() / 60
+                self.chart.evalJS(
+                    "Highcharts.setOptions({global: {timezoneOffset: %d}});"
+                    % -60
+                )  # Why is this negative? It works.
 
         self.chart.setXAxisType(
             'datetime'
@@ -551,6 +569,8 @@ class OWLineChart(OWWidget):
             if var.is_continuous and var != data.time_variable
         ]
         self.varmodel.wrap(variables)
+        self.set_attributes()
+
         self.update_plots()
 
     @Inputs.forecast
@@ -563,40 +583,41 @@ class OWLineChart(OWWidget):
 
     @Inputs.features
     def set_features(self, features: AttributeList) -> None:
-        if features and not self.features:
-            # if features are on the input and they were not before
-            # context should be saved for later when features will not be
-            # present anymore
-            self.closeContext()
+        self.closeContext()
+        # when features present selection is manipulated by  signal and not manually
+        self._disable_buttons(features is not None)
         self.features = features
         if self.data:
             self.set_attributes()
             self.update_plots()
 
+    def _disable_buttons(self, disabled):
+        """ Disable add plot button and buttons for removing the plot"""
+        self.add_button.setDisabled(disabled)
+        for c in self.configs:
+            c.button_close.setDisabled(disabled)
+
     def set_attributes(self) -> None:
         """
         In case when features present: set shown attributes to match features
         In case when features not present: set default value and open context
         """
-        self.closeContext()
+        features = None
         if self.features:
-            self.attrs = [
-                [f]
-                for f in self.features
+            features = [
+                f for f in self.features
                 if f in self.data.domain and f != self.data.time_variable
             ]
-            self.is_logit = [False] * len(self.attrs)
+            self.attrs = [[f] for f in features]
         else:
-            variables = [
-                var
-                for var in self.data.domain.variables
-                if var.is_continuous and var != self.data.time_variable
-            ]
-            self.attrs = [[variables[0]]]
-            self.is_logit = [False]
-            # context is only open when features not provided on input
-            # when provided features defines the selection
-            self.openContext(self.data.domain)
+            self.attrs = [self.varmodel[:1]]
+        self.is_logit = [False] * len(self.attrs)
+        self.plot_type = [PLOT_TYPE_ITEMS[0]] * len(self.attrs)
+        # update self.conf tp new setting - otherwise self.attrs will be set
+        # back to self.conf settings - openContext calls storeSpecificSettings
+        self.update_plots()
+
+        self.openContext((self.data.domain, features))
 
     def update_plots(self) -> None:
         """
@@ -613,39 +634,39 @@ class OWLineChart(OWWidget):
 
         assert len(self.configs) == len(self.attrs)
         assert len(self.attrs) == len(self.is_logit)
+        assert len(self.attrs) == len(self.plot_type)
         # select correct values
-        for config, attr, log in zip(self.configs, self.attrs, self.is_logit):
+        for config, attr, log, pt in zip(self.configs, self.attrs, self.is_logit, self.plot_type):
             config.set_logarithmic(log)
             config.set_selection(attr)
+            config.set_plot_type(pt)
             config.view.setEnabled(not self.features)
 
     def _selections_changed(self) -> None:
         for config in self.configs:
             config.selection_changed()
 
-    def closeEvent(self, event: QCloseEvent) -> None:
+    def storeSpecificSettings(self) -> None:
         """
-        When someone delets the widget closeContext must be called to gather
-        settings.
+        Gather configs in contextVariables before context is closed, widget is
+        closed or workflow is saved
         """
-        self.closeContext()
-        super().closeEvent(event)
+        super().storeSpecificSettings()
+        attrs, is_logit, plot_type = [], [], []
+        for config in self.configs:
+            attrs.append(config.get_selection())
+            is_logit.append(config.is_logarithmic)
+            plot_type.append(config.plot_type)
+        self.attrs = attrs
+        self.is_logit = is_logit
+        self.plot_type = plot_type
 
-    def closeContext(self) -> None:
-        """
-        Gather configs in contextVariables and close context.
-        """
-        if not self.features:
-            # only close in case of when features are not present if they are
-            # feature selection is defined by the input and context should
-            # not have impact
-            attrs, is_logit = [], []
-            for config in self.configs:
-                attrs.append(config.get_selection())
-                is_logit.append(config.is_logarithmic)
-            self.attrs = attrs
-            self.is_logit = is_logit
-            super().closeContext()
+    @classmethod
+    def migrate_context(cls, context: Context, version: int) -> None:
+        # all current context should have features attributes, if it is not
+        # present yet the context was made before the changes
+        if version < 2:
+            context.features = []
 
 
 if __name__ == "__main__":
